#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>

#define SIZE(a) ( sizeof( a ) / sizeof( a[0]) )
#define MASK(a) ( SIZE(a) - 1 )
#define CHECK_TAIL_HEAD(a) ( ((a)->position_head) == ((a)->position_tail) ) 


typedef struct
{
  uint32_t position_head;
  uint32_t position_tail;
  int arrm[ 1ULL << 4 ]; // for testing
} my_circ_buf;



inline void circ_buf_push_head (my_circ_buf **c_b, const register int a)
{
  (*c_b)->position_head &= MASK((*c_b)->arrm);

#ifdef DEBUG
  if (CHECK_TAIL_HEAD(*c_b))  fprintf(stderr, "head to tail overwrite\n");
#endif

  ((*c_b)->arrm) [(*c_b)->position_head] = a;

#ifdef DEBUG
  fprintf(stderr, "pos head : %d\nval : %d\n\n", (*c_b)->position_head, (*c_b)->arrm[(*c_b)->position_head]);
#endif

  (*c_b)->position_head++;
}

inline void circ_buf_push_tail (my_circ_buf **c_b, const register int a)
{
#ifdef DEBUG
  if (CHECK_TAIL_HEAD(*c_b))  fprintf(stderr, "tail to head overwrite\n");
#endif

  (*c_b)->position_tail--;  
  (*c_b)->position_tail &= MASK((*c_b)->arrm);
  ((*c_b)->arrm) [(*c_b)->position_tail] = a;

#ifdef DEBUG
  fprintf(stderr, "pos tail : %d\nval : %d\n\n", (*c_b)->position_tail, (*c_b)->arrm[(*c_b)->position_tail]);
#endif
}


int main (void)
{
  
  my_circ_buf *c_b = malloc(sizeof(my_circ_buf));
  c_b->position_head = 0;
  memset(c_b, 0x00, sizeof(c_b));
  /*
  for ( register int i = 0; i <= 9999999 ; i++ )
    {
    circ_buf_push_head(&c_b, i);
    }
  */

#define push_h(a) circ_buf_push_head(&c_b, a)
#define push_t(a) circ_buf_push_tail(&c_b, a)

  push_t(-1); push_t(-2); push_t(-3); push_t(-4); push_t(-5); push_t(-6); push_t(-7); push_t(-8); // push_t(-9);
  push_h( 1); push_h( 2); push_h( 3); push_h( 4); push_h( 5); push_h( 6); push_h( 7); push_h( 8); push_h( 9);
  //  push_t(-1); push_t(-2); push_t(-3); push_t(-4); push_t(-5); push_t(-6); push_t(-7); push_t(-8); push_t(-9);

#undef push_h
#undef push_t

  /*
  circ_buf_push_head(&c_b, 5); circ_buf_push_head(&c_b, 100);
  circ_buf_push_tail(&c_b,-5); circ_buf_push_tail(&c_b,-100);
  */  

  for (size_t i = 0; i < SIZE(c_b->arrm); i++ )
    printf("%d ", c_b->arrm[i]);

  printf("\n");
  fprintf( stderr,
           "mask : %X\n"
           "buf_size : %zu\n",
	   (uint8_t)MASK(c_b->arrm),
           SIZE(c_b->arrm)
         );

  free(c_b);

  return 0;
}